"""Base (non-end user) input models for qcio. Most models here are used to collect
and organize common attributes used across multiple input objects.
"""
from enum import Enum
from typing import Any, Dict, Optional

from pydantic import BaseModel

from .base_models import Files
from .molecule import Molecule

__all__ = ["Model", "CalcType", "ProgramArgs", "QCProgramArgs", "InputBase"]


class CalcType(str, Enum):
    """The Calculation type."""

    energy = "energy"
    gradient = "gradient"
    hessian = "hessian"
    optimization = "optimization"
    transition_state = "transition_state"

    def __repr__(self) -> str:
        """Custom repr for CalcType"""
        return f"'{self.name}'"  # pragma: no cover


class InputBase(Files):
    """Base class for all qcio input classes.

    Useful for type annotations on functions accepting any input object.
    """

    pass


class ProgramArgs(InputBase):
    """Core arguments for a program including keywords and files.

    Attributes:
        keywords: Dict of keywords to be passed to the program. Defaults to empty dict.

    """

    keywords: Dict[str, Any] = {}


class Model(BaseModel):
    """The model for the quantum chemistry calculation.

    Attributes:
        method: The name of the method to be used in the calculation. Named according to
            the convention of the program being called. If an MM calculation then the
            name of the force field.
        basis: The name of the basis set to be used in the calculation. Named according
            to the convention of the program being called.
    """

    method: str
    basis: Optional[str] = None


class QCProgramArgs(ProgramArgs):
    """Core arguments for a QC program. Breaks out model from keywords.

    Attributes:
        model: The model for the quantum chemistry calculation.
    """

    model: Model


class StructuredInputBase(ProgramArgs):
    """Base class for all structured qcio input classes.

    Structured inputs consist of args for a program, a Molecule, and a calctype.

    Attributes:
        molecule: The molecule to be used in the calculation.
        calctype: The type of calculation to be performed.
    """

    calctype: CalcType
    molecule: Molecule



from .base_models import *  # noqa: F403
from .inputs import *  # noqa: F403
from .inputs_base import *  # noqa: F403
from .molecule import *  # noqa: F403
from .outputs import *  # noqa: F403
from .outputs_base import *  # noqa: F403
from .results import *  # noqa: F403



"""End user input models for qcio"""
from pathlib import Path
from typing import List, Union

from typing_extensions import Self

from .inputs_base import InputBase, ProgramArgs, QCProgramArgs, StructuredInputBase
from .molecule import Molecule

__all__ = [
    "FileInput",
    "ProgramInput",
    "DualProgramInput",
]


class FileInput(InputBase):
    """File and command line argument inputs for a calculation.

    Attributes:
        files: A dict mapping filename to str or bytes data.
        cmdline_args: A list of command line arguments to be passed to the program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    cmdline_args: List[str] = []

    @classmethod
    def from_directory(cls, directory: Union[Path, str], **kwargs) -> Self:
        """Collect all files from directory and instantiate an object."""
        obj = cls(**kwargs)
        directory = Path(directory)
        obj.add_files(directory)
        return obj


class ProgramInput(StructuredInputBase, QCProgramArgs):
    """Input for a single point calculation.

    Attributes:
        calctype: The type of calculation to perform.
        model: The model for the quantum chemistry calculation.
        keywords: A dict of keywords to be passed to the program excluding model and
            calctype. Defaults to an empty dict.
        molecule: The molecule to be used in the calculation.
        files: Files to be passed to the QC program.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """


class DualProgramInput(StructuredInputBase, ProgramArgs):
    """Input for a two program calculation.

    Attributes:
        calctype: The type of calculation to be performed.
        keywords: Dict of keywords to be passed to the program. Defaults to empty dict.
        molecule: The molecule to be used in the calculation.
        files: A dict mapping filename to str or bytes data.
        subprogram_args: The input arguments for the subprogram.
        subprogram: The name of the subprogram to be used.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    subprogram_args: QCProgramArgs
    subprogram: str


if __name__ == "__main__":
    # Test geometry optimization input
    from qcio import CalcType, Model

    opt_inp = DualProgramInput(
        calctype=CalcType.optimization,
        keywords={"maxiter": 100},
        molecule=Molecule(
            symbols=["C", "C", "H", "H", "H", "H"],
            geometry=[
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 1.2],
                [0.0, 1.2, 0.0],
                [0.0, -1.2, 0.0],
                [1.2, 0.0, 0.0],
                [-1.2, 0.0, 0.0],
            ],
        ),
        subprogram_args=QCProgramArgs(
            keywords={"purify": "no"},
            model=Model(**{"method": "b3lyp", "basis": "6-31g"}),
        ),
        subprogram="terachem",
    )



"""Compatibility layer for QCElemental (QCSchema)."""
from typing import Any, Dict, Union

from qcio import (
    ProgramFailure,
    ProgramInput,
    Provenance,
    SinglePointOutput,
    SinglePointResults,
    Wavefunction,
)


def to_qcel_input(prog_input: ProgramInput) -> Dict[str, Any]:
    """Return the QCElemental v1 input schema representation of the input
    (AtomicInput dict).

    Args:
        prog_input: The input object to convert.

    Returns:
        The QCElemental v1 dict representation of an AtomicInput object.
    """
    return {
        "molecule": {
            "symbols": prog_input.molecule.symbols,
            "geometry": prog_input.molecule.geometry,
            "molecular_charge": prog_input.molecule.charge,
            "molecular_multiplicity": prog_input.molecule.multiplicity,
            "connectivity": prog_input.molecule.connectivity or None,
            "identifiers": {
                key: value
                for key, value in prog_input.molecule.identifiers.dict().items()
                if key not in ["name_IUPAC", "name_common"]  # not on qcel model
            },
        },
        "driver": prog_input.calctype,
        "model": prog_input.model.dict(),
        "keywords": prog_input.keywords,
        "extras": prog_input.extras,
    }


def from_qcel_output(
    qcio_input: ProgramInput,
    qcel_output: Dict[str, Any],
) -> Union[SinglePointOutput, ProgramFailure]:
    """Create a SinglePointSuccessfulOutput or SinglePointFailedOutput from the
    QCElemental v1 output schema representation of the output (AtomicResult dict).

    NOTE: Function no longer used in qcop; only using from_qcel_output_results. Maybe
        remove in the future if I never return to use it.

    Args:
        qcio_input: The input object for the computation.
        qcel_output: The QCElemental v1 output schema representation of the output.
            May be a dict representing an AtomicResult or FailedOperation.
    """
    prov_extras = {"NOTE": "Computed by QCEngine"}

    if qcel_output["success"] is False:
        return ProgramFailure(
            input_data=qcio_input,
            traceback=qcel_output["error"]["error_message"],
            provenance=Provenance(
                program=qcel_output["input_data"]["provenance"]["creator"],
                program_version=qcel_output["input_data"]["provenance"].get("version"),
                wall_time=qcel_output["input_data"]["provenance"].get("wall_time"),
                extras=prov_extras,
            ),
            extras=qcel_output["extras"] or {},  # Because .extras may be None
        )
    else:
        return SinglePointOutput(
            input_data=qcio_input,
            stdout=qcel_output["stdout"],
            results=from_qcel_output_results(qcel_output),
            provenance=Provenance(
                program=qcel_output["provenance"]["creator"],
                program_version=qcel_output["provenance"].get("version"),
                wall_time=qcel_output["provenance"].get("wall_time"),
                extras=prov_extras,
            ),
        )


def from_qcel_output_results(
    qcel_output: Dict[str, Any],
) -> SinglePointResults:
    """Create a SinglePointSuccessfulOutput or SinglePointFailedOutput from the
    QCElemental v1 output schema representation of the output (AtomicResult dict).

    Args:
        qcel_output: The QCElemental v1 output schema representation of the output.
            May be a dict representing an AtomicResult or FailedOperation.
    """
    # Collect values from keys that exist in qcio
    qcio_to_qcel = {
        "calcinfo_natoms": "calcinfo_natom",
        "energy": "return_energy",
        "gradient": "return_gradient",
        "hessian": "return_hessian",
    }
    results = {}
    for key in SinglePointResults.__annotations__:
        if key in qcio_to_qcel:
            qcel_key = qcio_to_qcel[key]
        else:
            qcel_key = key
        value = qcel_output["properties"].get(qcel_key)
        if value is not None:
            results[key] = value

    # Override with return_result as qcel may not have save the key value to .properties
    results[qcel_output["driver"]] = qcel_output["return_result"]

    if qcel_output["wavefunction"]:
        results["wavefunction"] = {
            key: value
            for key, value in qcel_output["wavefunction"].items()
            if key in Wavefunction.__annotations__
        }

    results["extras"] = {"extras": {"NOTE": "Results computed using QCEngine"}}
    return SinglePointResults(**results)



"""Objects containing computed result values from a calculation."""
from __future__ import annotations

from typing import List, Optional

import numpy as np
from pydantic import validator

from qcio.helper_types import ArrayLike2D, ArrayLike3D

from .base_models import QCIOModelBase
from .molecule import Molecule
from .outputs_base import ResultsBase

__all__ = ["SinglePointResults", "Wavefunction", "OptimizationResults"]


class Wavefunction(QCIOModelBase):
    """The wavefunction for a single point calculation.

    Attributes:
        scf_eigenvalues_a: The SCF alpha-spin orbital eigenvalues.
        scf_eigenvalues_b: The SCF beta-spin orbital eigenvalues.
        scf_occupations_a: The SCF alpha-spin orbital occupations.
        scf_occupations_b: The SCF beta-spin orbital occupations.
    """

    scf_eigenvalues_a: Optional[ArrayLike2D] = None
    scf_eigenvalues_b: Optional[ArrayLike2D] = None
    scf_occupations_a: Optional[ArrayLike2D] = None
    scf_occupations_b: Optional[ArrayLike2D] = None

    _to_numpy = validator(
        "scf_eigenvalues_a",
        "scf_eigenvalues_b",
        "scf_occupations_a",
        "scf_occupations_b",
        allow_reuse=True,
    )(lambda x: np.asarray(x) if x is not None else None)


class SinglePointResults(ResultsBase):
    """The computed results from a single point calculation.

    Attributes:
        calcinfo_natoms: The number of atoms as computed by the program.
        calcinfo_nalpha: The number of alpha electrons as computed by the program.
        calcinfo_nbeta: The number of beta electrons as computed by the program.
        calcinfo_nbasis: The number of basis functions in the calculation.
        calcinfo_nmo: The number of molecular orbitals in the calculation

        energy: The electronic energy of the molecule in Hartrees.
        gradient: The gradient of the molecule in Hartrees/Bohr.
        hessian: The hessian of the molecule in Hartrees/Bohr^2.
        nuclear_repulsion_energy: The nuclear repulsion energy of the molecule in
            Hartrees.

        wavefunction: Wavefunction data from the calculation.

        freqs_wavenumber: The frequencies of the molecule in wavenumbers.
        normal_modes_cartesian: n_vibmodes by n_atoms length array containing
            un-mass-weighted Cartesian displacements of each normal mode.
        gibbs_free_energy: Gibbs free energy (i.e. thermochemical analysis) in Hartrees
            of a system where translation / rotation / vibration degrees of freedom are
            approximated using ideal gas / rigid rotor / harmonic oscillator
            respectively.
    """

    # calcinfo contains general information about the calculation
    calcinfo_natoms: Optional[int] = None
    calcinfo_nbasis: Optional[int] = None
    calcinfo_nmo: Optional[int] = None
    calcinfo_nalpha: Optional[int] = None
    calcinfo_nbeta: Optional[int] = None

    # Core properties
    energy: Optional[float] = None
    gradient: Optional[ArrayLike2D] = None
    hessian: Optional[ArrayLike2D] = None
    nuclear_repulsion_energy: Optional[float] = None

    # Wavefunction data
    wavefunction: Optional[Wavefunction] = None

    # Frequency data
    freqs_wavenumber: List[float] = []
    normal_modes_cartesian: Optional[ArrayLike3D] = None
    gibbs_free_energy: Optional[float] = None

    @validator("normal_modes_cartesian")
    def validate_normal_modes_cartesian_shape(cls, v: ArrayLike3D):
        if v is not None:
            # Assume array has length of the number of normal modes
            n_normal_modes = len(v)
            return np.asarray(v).reshape(n_normal_modes, -1, 3)

    @validator("gradient")
    def validate_gradient_shape(cls, v: ArrayLike2D):
        """Validate gradient is n x 3"""
        if v is not None:
            return np.asarray(v).reshape(-1, 3)

    @validator("hessian")
    def validate_hessian_shape(cls, v: ArrayLike2D):
        """Validate hessian is square"""
        if v is not None:
            v = np.asarray(v)
            n = int(np.sqrt(v.size))
            return v.reshape((n, n))


class OptimizationResults(ResultsBase):
    """Computed properties for an optimization.

    Attributes:
        energies: The energies for each step of the optimization.
        # NOTE: Maybe should just be be geometries since all other data is duplicated?
        molecules: The Molecule objects for each step of the optimization. Must contain
            at least one Molecule corresponding to the final geometry.
        sp_outputs: The SinglePointOutput objects for each step of the optimization.
    """

    energies: List[float]
    molecules: List[Molecule]
    # NOTE: This is a circular import, hence the ignores.
    sp_outputs: List["SinglePointOutput"] = []  # type: ignore  # noqa: F821



"""Base (non end user) classes for output models."""
from typing import TYPE_CHECKING, Literal, Optional

from .base_models import Files, Provenance, QCIOModelBase
from .inputs_base import InputBase

if TYPE_CHECKING:  # pragma: no cover
    from pydantic.typing import ReprArgs


__all__ = [
    "ResultsBase",
    "OutputBase",
]


class ResultsBase(QCIOModelBase):
    """Base class for all results computed by a program.

    Currently this is just a placeholder class. Perhaps some day Results objects will
    have some common interface. Helpful for generic type annotations in qcop.
    """

    pass


class OutputBase(Files):
    """Base class for all output classes.

    Attributes:
        input_data: The input object for the computation.
        success: A boolean indicator that the operation succeeded or failed. Allows
            programmatic assessment of all outputs regardless of if they failed or
            succeeded by checking `output.success`.
        results: An object containing the results computed by the program.
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
        provenance: An object containing the provenance information for the output.
    """

    input_data: InputBase
    success: bool
    stdout: Optional[str] = None
    results: ResultsBase = ResultsBase()
    provenance: Provenance

    @property
    def pstdout(self) -> None:
        """Print the stdout text"""
        print(self.stdout)

    def __repr_args__(self) -> "ReprArgs":
        """Exclude stdout from the repr"""
        return [
            (key, value if key != "stdout" else "<...>")
            for key, value in super().__repr_args__()
        ]


class SuccessfulOutputBase(OutputBase):
    """Base object for any successful computation.

    Attributes:
        success: Always True for a successful output.
    """

    success: Literal[True] = True



"""End user output objects from a calculation."""
from typing import Literal, Optional, Union

from qcio.helper_types import ArrayLike2D

from .inputs import DualProgramInput, FileInput, ProgramInput
from .molecule import Molecule
from .outputs_base import OutputBase, SuccessfulOutputBase
from .results import OptimizationResults, SinglePointResults

__all__ = ["FileOutput", "SinglePointOutput", "OptimizationOutput", "ProgramFailure"]


class FileOutput(SuccessfulOutputBase):
    """Generic output class for pure file based I/O.

    Attributes:
        input_data: The FileInput object used for the computation.
        files: A dict mapping filename to str or bytes data.
        success: A boolean indicator that the operation succeeded (always True)
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
        provenance: An object containing the provenance information for the results.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.

    """

    input_data: FileInput


class SinglePointOutput(SuccessfulOutputBase):
    """Output from a successful single point calculation.

    Attributes:
        input_data: The SinglePointInput object for the computation.
        success: Always True for a successful computation.
        results: The results computed by the program.
        files: A dict mapping filename to str or bytes data.
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
        provenance: An object containing the provenance information for the results.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    input_data: ProgramInput
    results: SinglePointResults

    @property
    def return_result(self) -> Union[float, ArrayLike2D]:
        """Return the result of the calculation.

        Returns:
            The explicitly requested result of the calculation, i.e., the energy,
                gradient, or hessian.
        """
        return getattr(self.results, self.input_data.calctype.value)


class OptimizationOutput(SuccessfulOutputBase):
    """Output from a successful optimization.

    Attributes:
        input_data: The OptimizationInput object for the computation.
        results: The results computed by the program.
    """

    input_data: Union[DualProgramInput, ProgramInput]
    results: OptimizationResults

    @property
    def final_molecule(self) -> Molecule:
        """Return the final molecule in the optimization."""
        return self.results.molecules[-1]


class ProgramFailure(OutputBase):
    """A object containing details about a failed calculation.

    Attributes:
        input_data: The input object for the computation.
        success: Always False for a Failed output.
        traceback: String representation of the traceback of the exception that caused
            the failure.
        results: Any compted data that was able to be extracted before program failed.
        stdout: The primary logging output of the program. Contains a union of stdout
            and stderr.
        provenance: An object containing the provenance information for the output.
    """

    input_data: Union[DualProgramInput, ProgramInput, FileInput]
    success: Literal[False] = False
    traceback: Optional[str] = None

    @property
    def ptraceback(self) -> None:
        """Print the traceback text"""
        print(self.traceback)



"""The Base model from which all QCIO Model objects inherit."""
import json
from abc import ABC
from base64 import b64decode, b64encode
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import numpy as np
import toml
import yaml
from pydantic import BaseModel, Extra, validator
from typing_extensions import Self

from ..helper_types import StrOrPath

if TYPE_CHECKING:  # pragma: no cover
    from pydantic.typing import ReprArgs


__all__ = ["Files", "Provenance"]


class QCIOModelBase(BaseModel, ABC):
    """Base Model for all QCIO objects.

    Attributes:
        version: The version of the schema.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    # NOTE: No version for now until we have a stable schema
    # Or maybe have version be version of qcio that generated the object?
    # version: ClassVar[Literal["v1"]] = "v1"
    extras: Dict[str, Any] = {}

    class Config:
        # Raises an error if extra fields are passed to model.
        extra = Extra.forbid
        # Allow numpy types in models. Pydantic will no longer raise an exception for
        # types it doesn't recognize.
        # https://docs.pydantic.dev/latest/usage/types/#arbitrary-types-allowed
        arbitrary_types_allowed = True
        # Don't allow mutation of objects
        # https://docs.pydantic.dev/latest/usage/models/#faux-immutability
        allow_mutation = False
        # convert ndarray to list for JSON serialization
        # https://docs.pydantic.dev/latest/usage/exporting_models/#json_encodershttps://docs.pydantic.dev/latest/usage/exporting_models/#json_encoders  # noqa: E501
        json_encoders = {np.ndarray: lambda v: v.tolist()}
        # exclude fields with value None from serialization
        exclude_none = True

    def dict(self, **kwargs):
        """Convert the object to a dictionary.

        Properly serialize numpy arrays. Serialization is performed in .dict() so that
        multiple string serializers can used it without duplicating logic
        (e.g. json, yaml, toml).
        """
        model_dict = super().dict(**kwargs)
        to_pop = []
        for key, value in model_dict.items():
            # Custom serialization for numpy arrays, enums, and pathlib Paths
            if isinstance(value, np.ndarray):
                model_dict[key] = value.tolist()
            elif issubclass(type(value), Enum):
                model_dict[key] = value.value
            elif isinstance(value, Path):
                model_dict[key] = str(value)

            # Exclude empty lists, dictionaries, and objects with all None values from
            # serialization
            elif value in [None, [], {}]:
                to_pop.append(key)

        for key in to_pop:
            model_dict.pop(key)

        return model_dict

    @classmethod
    def open(cls, filepath: Union[Path, str]) -> Self:
        """Instantiate an object from data saved to disk.

        Args:
            filepath: The path to the object on disk.
            *args: Additional arguments to pass to the constructor.
            **kwargs: Additional keyword arguments to pass to the constructor.

        Returns:
            The instantiated object.
        """
        filepath = Path(filepath)
        data = filepath.read_text()

        if filepath.suffix in [".yaml", ".yml"]:
            return cls.parse_obj(yaml.safe_load(data))
        elif filepath.suffix == ".toml":
            return cls.parse_obj(toml.loads(data))

        # Assume json for all other file extensions
        return cls.parse_raw(data)
        # pydantic v2
        # return cls.model_validate_json(filepath.read_text())

    def save(self, filepath: Union[Path, str], **kwargs) -> None:
        """Write an object to disk as json.

        Args:
            filepath: The path to write the object to.I
        """
        filepath = Path(filepath)
        filepath.parent.mkdir(exist_ok=True, parents=True)

        if filepath.suffix in [".yaml", ".yml"]:
            data = yaml.dump(self.dict(**kwargs))

        elif filepath.suffix == ".toml":
            data = toml.dumps(self.dict(**kwargs))

        else:
            # Write data to json regardless of file extension
            data = self.json(**kwargs)

        filepath.write_text(data)
        # pydantic v2
        # filepath.write(self.model_dump())

    def __repr_args__(self) -> "ReprArgs":
        """Only show non empty fields in repr."""

        def exists(value):
            if isinstance(value, np.ndarray):
                return value.size > 0
            return bool(value)

        return [  # pragma: no cover
            (name, value) for name, value in self.__dict__.items() if exists(value)
        ]


class Files(QCIOModelBase):
    """File model for handling string and binary data.

    Binary data is encoded as base64 strings during serialization.

    Attributes:
        files: A dict mapping filename to str or bytes data.
    """

    files: Dict[str, Union[str, bytes]] = {}

    @validator("files", pre=True)
    def convert_base64_to_bytes(cls, value):
        """Convert base64 encoded data to bytes."""
        for filename, data in value.items():
            if isinstance(data, str) and data.startswith("base64:"):
                value[filename] = b64decode(data[7:])
        return value

    def dict(self, *args, **kwargs):
        """Return a dict representation of the object encoding bytes as b64 strings."""
        dict = super().dict(*args, **kwargs)
        if self.files:  # clause so that empty files dict is not included in dict
            files = {}
            for filename, data in self.files.items():
                if isinstance(data, bytes):
                    data = f"base64:{b64encode(data).decode('utf-8')}"
                files[filename] = data
            dict["files"] = files
        return dict

    def json(self, *args, **kwargs):
        """Return a JSON representation of the object."""
        return json.dumps(self.dict(*args, **kwargs))

    def add_file(
        self, filepath: Union[Path, str], relative_dir: Optional[Path] = None
    ) -> None:
        """Create a File object from a file on disk.

        Args:
            filepath: The path to the file.
            relative_dir: The directory to make the file relative to. Helpful when
                adding files from a subdirectory.
        """
        filepath = Path(filepath)
        raw_bytes = filepath.read_bytes()
        try:
            data: Union[str, bytes] = raw_bytes.decode("utf-8")  # str
        except UnicodeDecodeError:
            data = raw_bytes  # bytes

        # Set filename relative to relative_dir
        if relative_dir:
            filename = str(filepath.relative_to(relative_dir))
        else:
            filename = filepath.name

        self.files[filename] = data

    def add_files(
        self,
        directory: StrOrPath,
        recursive: bool = False,
        exclude: Optional[List[str]] = None,
    ) -> None:
        """Add all files in a directory to the object.

        Args:
            directory: The directory to add files from.
            recursive: Whether to recursively add files from subdirectories.
            exclude: A list of filenames to exclude from the directory.
        """
        directory = Path(directory)
        if exclude is None:
            exclude = []
        if recursive:
            files = directory.rglob("*")
        else:
            files = directory.glob("*")
        for filepath in files:
            if filepath.is_file() and filepath.name not in exclude:
                self.add_file(filepath, directory)

    def write_files(self, directory: StrOrPath = Path(".")) -> None:
        """Write all files to the specified directory"""
        directory = Path(directory)
        directory.mkdir(exist_ok=True)
        for filename, data in self.files.items():
            mode = "w" if isinstance(data, str) else "wb"
            filepath = directory / filename
            # In case filename is a relative path, create the parent directories
            filepath.parent.mkdir(exist_ok=True, parents=True)
            filepath.open(mode).write(data)

    def __repr_args__(self) -> "ReprArgs":
        """Replace file data with '<bytes>' or '<str>' in __repr__."""
        rargs = super().__repr_args__()
        return [
            (k, v)
            if k != "files"
            else (
                k,
                {
                    name: bytes if isinstance(data, bytes) else str
                    for name, data in v.items()
                },
            )
            for k, v in rargs
        ]


class Provenance(QCIOModelBase):
    """Provenance information for a QC program.

    Attributes:
        program: The name of the program that created the output.
        version: The version of the program that created the output.
        working_dir: The working directory used by the program.
        wall_time: The wall time used by the program.
        hostname: The hostname of the machine the program was run on.
        hostcpus: The number of logical cpus on the host machine where the program ran.
        hostmem: The amount of memory on the host machine where the program ran in GiB.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.

    """

    program: str
    program_version: Optional[str] = None
    working_dir: Optional[Path] = None
    wall_time: Optional[float] = None
    hostname: Optional[str] = None
    hostcpus: Optional[int] = None
    hostmem: Optional[int] = None



from typing import Optional

from .models import (
    CalcType,
    FileOutput,
    Molecule,
    OptimizationOutput,
    SinglePointOutput,
)

# Helper Molecules
water = Molecule(
    symbols=["O", "H", "H"],
    geometry=[0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0, 1.0, 0.0],
    charge=0,
    multiplicity=1,
    connectivity=[(0, 1, 1.0), (0, 2, 1.0)],
)


def calctype_to_output(calctype: Optional[CalcType]):
    """Return the output object for the given calculation type."""
    calctype_to_output = {
        None: FileOutput,  # Covers FileInput case
        CalcType.energy: SinglePointOutput,
        CalcType.gradient: SinglePointOutput,
        CalcType.hessian: SinglePointOutput,
        CalcType.optimization: OptimizationOutput,
        CalcType.transition_state: OptimizationOutput,
    }
    return calctype_to_output[calctype]
